package com.tidecloud.dataacceptance.service;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.annotation.PostConstruct;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import com.tidecloud.dataacceptance.config.AcceptanceDeviceConfig;
import com.tidecloud.dataacceptance.config.DeviceConfig;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.timeout.IdleStateHandler;

/**
 * Created by ryan
 */
@Service
public class AcceptanceDeviceService implements ApplicationContextAware {
	private static Logger logger = LoggerFactory.getLogger(AcceptanceDeviceService.class);
	private ApplicationContext context;

	@Autowired
	private AcceptanceDeviceConfig acceptanceDeviceListConfig;

	@PostConstruct
	public void transfer() throws IOException, ParseException {

		List<DeviceConfig> deviceListConfig = acceptanceDeviceListConfig.getDeviceList();
		List<DeviceConfig> workDeviceList = new ArrayList<>();
		for (DeviceConfig config : deviceListConfig) {
			if (config.isEnable()) {
				workDeviceList.add(config);
			}
		}

		if (workDeviceList.size() > 0) {
			ExecutorService singleThreadPool = new ThreadPoolExecutor(workDeviceList.size(), 20, 1L,
					TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>(1024), Executors.defaultThreadFactory(),
					new java.util.concurrent.ThreadPoolExecutor.AbortPolicy());

			for (int i = 0; i < workDeviceList.size(); i++) {
				DeviceConfig config = workDeviceList.get(i);
				singleThreadPool.execute(() -> {
					MDC.put("strategyName", config.getName());
					try {
						Class handlerClass = Class.forName(config.getHandlerClass());
						AcceptanceInboundHandlerAdapter handler = (AcceptanceInboundHandlerAdapter) context
								.getBean(handlerClass);
						handler.setDataPath(config.getDataFileDir());
						handler.setTopic(config.getTopic());
						String ip = getIp("eth0");
						handler.setIp(ip);
						handler.setPort(config.getPort());
						handler.setPrefixName(config.getName());
						start(handler);
					} catch (ClassNotFoundException e) {
						logger.error(e.getMessage());
					} catch (Exception ex) {
						logger.error(ex.getMessage());
					}

				});
			}
		}

	}

	public static String getIp(String eth0) {
		String hostAddress = "127.0.0.1";
		try {
			Enumeration<NetworkInterface> interfaces = null;
			interfaces = NetworkInterface.getNetworkInterfaces();
			while (interfaces.hasMoreElements()) {
				NetworkInterface ni = interfaces.nextElement();
				if (eth0.equalsIgnoreCase(ni.getName())) {
					Enumeration<InetAddress> addresss = ni.getInetAddresses();
					while (addresss.hasMoreElements()) {
						InetAddress ip = addresss.nextElement();
						if (ip != null && ip instanceof Inet4Address) {
							if (ip.getHostAddress().equals("127.0.0.1")) {
								continue;
							}
							return	hostAddress = ip.getHostAddress();
						}
					}
				}
			}
		} catch (Exception e) {
			logger.error(e.getMessage());
		}
		return hostAddress;
	}

	public void start(AcceptanceInboundHandlerAdapter handler) {
		EventLoopGroup bossGroup = new NioEventLoopGroup();
		EventLoopGroup workerGroup = new NioEventLoopGroup();
		try {
			ServerBootstrap b = new ServerBootstrap();
			b.group(bossGroup, workerGroup).channel(NioServerSocketChannel.class)
					.childHandler(new ChannelInitializer<SocketChannel>() {
						@Override
						protected void initChannel(SocketChannel ch) throws Exception {
							ch.pipeline().addLast("idleStateHandler", new IdleStateHandler(10, 0, 0, TimeUnit.MINUTES));
							ch.pipeline().addLast(handler);
						}
					}).option(ChannelOption.SO_BACKLOG, 128).childOption(ChannelOption.SO_KEEPALIVE, true);

			ChannelFuture f = b.bind(handler.getPort()).sync();
			logger.info("start accept service for {}, bind address {}:{}", handler.getPrefixName(), handler.getIp(),
					handler.getPort());
			f.channel().closeFuture().sync();
		} catch (Exception e) {
			logger.error(e.getMessage());
		} finally {
			workerGroup.shutdownGracefully();
			bossGroup.shutdownGracefully();
		}

	}

	@Override
	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
		this.context = applicationContext;
	}

}
